"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.qTreeVec = exports.BoundsVec = void 0;
const vector_1 = require("@4bitlabs/vector");
const bounds_1 = require("./bounds");
class BoundsVec {
    constructor(boundsFn, initialCapacity) {
        this.boundsFn = boundsFn;
        this.vector = new vector_1.Vector(Float64Array, initialCapacity);
        this.indexes = new WeakMap();
    }
    add(item) {
        const bounds = this.boundsFn(item);
        const idx = this.vector.length;
        this.indexes.set(item, idx);
        this.vector.pushN(bounds);
        return bounds;
    }
    get(item) {
        var _a;
        const idx = (_a = this.indexes.get(item)) !== null && _a !== void 0 ? _a : null;
        if (idx === null) {
            const bounds = this.boundsFn(item);
            console.warn(`unexpected bounds cache miss!`);
            return bounds;
        }
        return this.vector.subarray(idx, idx + 4);
    }
    delete(item) {
        this.indexes.delete(item);
    }
    clear() {
        this.vector.clear();
    }
}
exports.BoundsVec = BoundsVec;
class qTreeVec {
    constructor([x0, y0, x1, y1], options, bounds, quadCache, depth = 0) {
        const { maxDepth = 7, maxChildren = 10 } = options;
        this.maxDepth = maxDepth;
        this.maxChildren = maxChildren;
        this.bounds = bounds;
        this.quadCache = quadCache;
        this.depth = depth;
        const midX = (x0 + x1) / 2;
        const midY = (y0 + y1) / 2;
        this.childAreas = [
            [x0, y0, midX, midY],
            [midX, y0, x1, midY],
            [x0, midY, midX, y1],
            [midX, midY, x1, y1],
        ];
        this.isSplit = false;
        this.children = [null, null, null, null];
        this.items = new Set();
    }
    clear() {
        this.items.clear();
        this.bounds.clear();
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.clear();
            this.children[i] = null;
        }
    }
    insert(item) {
        const bounds = this.bounds.add(item);
        this._insert(item, bounds);
    }
    remove(item) {
        if (!this.items.has(item)) {
            const quad = this.quadCache.get(item);
            if (!quad)
                return false;
            return quad.remove(item);
        }
        this._remove(item);
        return true;
    }
    collectAll(result = []) {
        for (const r of this.items)
            result.push(r);
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            child.collectAll(result);
        }
        return result;
    }
    collect(area, result = []) {
        for (const item of this.items) {
            const rect = this.bounds.get(item);
            if ((0, bounds_1.overlaps)(area, rect))
                result.push(item);
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                child.collectAll(result);
            else if ((0, bounds_1.overlaps)(area, childArea))
                child.collect(area, result);
        }
        return result;
    }
    *all() {
        yield* this.items;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            yield* child.all();
        }
    }
    *search(area) {
        for (const item of this.items) {
            const rect = this.bounds.get(item);
            if ((0, bounds_1.overlaps)(area, rect))
                yield item;
        }
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            const childArea = this.childAreas[i];
            if ((0, bounds_1.contains)(area, childArea))
                yield* child.all();
            else if ((0, bounds_1.overlaps)(area, childArea))
                yield* child.search(area);
        }
    }
    size() {
        let count = this.items.size;
        for (let i = 0; i < 4; i++) {
            const child = this.children[i];
            if (!child)
                continue;
            count += child.size();
        }
        return count;
    }
    // Internal Members
    /** @internal */
    _remove(item) {
        this.items.delete(item);
        this.bounds.delete(item);
        this.quadCache.delete(item);
    }
    _insert(item, bounds) {
        const isBottom = this.depth + 1 >= this.maxDepth;
        const isFull = this.items.size + 1 > this.maxChildren;
        if (!this.isSplit && !isBottom && isFull) {
            this.isSplit = true;
            for (const i of this.items) {
                const itemBounds = this.bounds.get(i);
                if (this._tryInsertChild(i, itemBounds)) {
                    this.items.delete(item);
                }
            }
        }
        if (this.isSplit && !isBottom) {
            if (this._tryInsertChild(item, bounds))
                return;
        }
        this.items.add(item);
        this.quadCache.set(item, this);
    }
    // Private Members
    _tryInsertChild(item, bounds) {
        for (let i = 0; i < 4; i++) {
            const childRect = this.childAreas[i];
            if ((0, bounds_1.contains)(childRect, bounds)) {
                let child = this.children[i];
                if (!child) {
                    this.children[i] = child = new qTreeVec(childRect, { maxDepth: this.maxDepth }, this.bounds, this.quadCache, this.depth + 1);
                }
                child._insert(item, bounds);
                return true;
            }
        }
        return false;
    }
}
exports.qTreeVec = qTreeVec;
//# sourceMappingURL=q-tree-vec.js.map